from __future__ import annotations

import abc
import inspect
import typing

from typing_extensions import get_original_bases

__all__ = ['is_abstract', 'resolve_generic_args']


def is_abstract(cls: type) -> bool:
    return inspect.isabstract(cls) or abc.ABC in cls.__bases__


def resolve_generic_args(cls: type, base_class: type) -> tuple[type, ...] | None:
    """Walk the MRO and return the first set of concrete generic arguments bound to *base_class*."""
    for klass in cls.__mro__:
        for base in get_original_bases(klass):
            origin = typing.get_origin(base)
            if origin is None or not isinstance(origin, type):  # pragma: no cover
                continue
            try:
                is_match = issubclass(origin, base_class)
            except TypeError:  # pragma: no cover
                continue
            if not is_match:
                continue
            args = typing.get_args(base)
            if args and isinstance(args[0], type):
                return args
    return None



from __future__ import annotations

from waku.eventsourcing.projection.lock.sqlalchemy.advisory import PostgresAdvisoryProjectionLock
from waku.eventsourcing.projection.lock.sqlalchemy.lock import PostgresLeaseProjectionLock
from waku.eventsourcing.projection.lock.sqlalchemy.tables import bind_lease_tables

__all__ = [
    'PostgresAdvisoryProjectionLock',
    'PostgresLeaseProjectionLock',
    'bind_lease_tables',
]



from waku.eventsourcing.serialization._retort import default_retort
from waku.eventsourcing.serialization.interfaces import IEventSerializer, ISnapshotStateSerializer
from waku.eventsourcing.serialization.json import JsonEventSerializer, JsonSnapshotStateSerializer
from waku.eventsourcing.serialization.registry import EventTypeRegistry

__all__ = [
    'EventTypeRegistry',
    'IEventSerializer',
    'ISnapshotStateSerializer',
    'JsonEventSerializer',
    'JsonSnapshotStateSerializer',
    'default_retort',
]



import inspect
from collections.abc import Callable, Sequence
from typing import Any, get_type_hints

from dishka import Provider, Scope
from dishka.entities.marker import BaseMarker

__all__ = [
    'activator',
    'contextual',
    'many',
    'object_',
    'provider',
    'scoped',
    'singleton',
    'transient',
]


def activator(fn: Callable[..., bool], *markers: Any) -> Provider:
    """Create a Provider with an activator for simple cases.

    Args:
        fn: Callable that returns bool to determine marker activation.
        *markers: Marker instances or types to activate.

    Returns:
        Provider with the activator registered.
    """
    p = Provider()
    p.activate(fn, *markers)
    return p


def provider(
    source: Callable[..., Any] | type[Any],
    *,
    scope: Scope = Scope.REQUEST,
    provided_type: Any | None = None,
    cache: bool = True,
    when: BaseMarker | None = None,
) -> Provider:
    provider_ = Provider(scope=scope)
    provider_.provide(source, provides=provided_type, cache=cache, when=when)
    return provider_


def singleton(
    interface_or_source: type[Any] | Callable[..., Any],
    implementation: type[Any] | Callable[..., Any] | None = None,
    /,
    *,
    when: BaseMarker | None = None,
) -> Provider:
    """Create a singleton provider (lifetime: app).

    Args:
        interface_or_source: Interface type or source if no separate implementation.
        implementation: Implementation type if interface is provided.
        when: Optional marker to conditionally activate the provider.

    Returns:
        Provider configured for singleton scope.
    """
    if implementation is not None:
        return provider(implementation, scope=Scope.APP, provided_type=interface_or_source, when=when)
    return provider(interface_or_source, scope=Scope.APP, when=when)


def scoped(
    interface_or_source: type[Any] | Callable[..., Any],
    implementation: type[Any] | Callable[..., Any] | None = None,
    /,
    *,
    when: BaseMarker | None = None,
) -> Provider:
    """Create a scoped provider (lifetime: request).

    Args:
        interface_or_source: Interface type or source if no separate implementation.
        implementation: Implementation type if interface is provided.
        when: Optional marker to conditionally activate the provider.

    Returns:
        Provider configured for request scope.
    """
    if implementation is not None:
        return provider(implementation, scope=Scope.REQUEST, provided_type=interface_or_source, when=when)
    return provider(interface_or_source, scope=Scope.REQUEST, when=when)


def transient(
    interface_or_source: type[Any] | Callable[..., Any],
    implementation: type[Any] | Callable[..., Any] | None = None,
    /,
    *,
    when: BaseMarker | None = None,
) -> Provider:
    """Create a transient provider (new instance per injection).

    Args:
        interface_or_source: Interface type or source if no separate implementation.
        implementation: Implementation type if interface is provided.
        when: Optional marker to conditionally activate the provider.

    Returns:
        Provider configured for transient (no cache) scope.
    """
    if implementation is not None:
        return provider(implementation, scope=Scope.REQUEST, provided_type=interface_or_source, cache=False, when=when)
    return provider(interface_or_source, scope=Scope.REQUEST, cache=False, when=when)


def object_(
    obj: Any,
    *,
    provided_type: Any | None = None,
    when: BaseMarker | None = None,
) -> Provider:
    """Provide the exact object passed at creation time as a singleton dependency.

    Args:
        obj: The instance to provide as-is.
        provided_type: Explicit type to provide (default: inferred).
        when: Optional marker to conditionally activate the provider.

    Returns:
        Provider configured to return the given object.
    """
    actual_type = provided_type if provided_type is not None else type(obj)
    return provider(lambda: obj, scope=Scope.APP, provided_type=actual_type, cache=True, when=when)


def contextual(
    provided_type: Any,
    *,
    scope: Scope = Scope.REQUEST,
) -> Provider:
    """Provide a dependency from the current context (e.g., app/request).

    Args:
        provided_type: The type to resolve from context.
        scope: Scope of the context variable (default: Scope.REQUEST).

    Returns:
        Provider configured for context resolution.
    """
    provider_ = Provider()
    provider_.from_context(provided_type, scope=scope)
    return provider_


def _get_provided_type(impl: Any) -> Any:
    if inspect.isclass(impl):
        return impl

    if callable(impl):
        hints = get_type_hints(impl)
        return_type = hints.get('return')
        if return_type is None:
            name = getattr(impl, '__name__', repr(impl))
            msg = f"Factory function '{name}' must have a return type annotation"
            raise TypeError(msg)
        return return_type

    msg = f'Implementation must be a class or callable, got {type(impl).__name__}'
    raise TypeError(msg)


def many(
    interface: Any,
    *implementations: Any,
    scope: Scope = Scope.REQUEST,
    cache: bool = True,
    when: BaseMarker | None = None,
    collect: bool = True,
) -> Provider:
    """Register multiple implementations as a collection.

    Args:
        interface: Interface type for the collection.
        *implementations: Implementation types or factory functions to include in collection.
        scope: Scope of the collection (default: Scope.REQUEST).
        cache: Whether to cache the resolve results within scope.
        when: Optional marker to conditionally activate the provider.
        collect: Whether to include collect+alias for Sequence/list resolution.
            Set to False to only register implementations without the collector.

    Returns:
        Provider configured for collection resolution.

    Raises:
        ValueError: If no implementations and collect is False.
    """
    if not implementations and not collect:
        msg = 'At least one implementation must be provided when collect=False'
        raise ValueError(msg)

    provider_ = Provider(scope=scope)
    for impl in implementations:
        provider_.provide(impl, provides=interface, cache=cache, when=when)
    if collect:
        provider_.collect(interface, scope=scope, cache=cache, provides=Sequence[interface])
        provider_.alias(Sequence[interface], provides=list[interface], cache=cache)
    return provider_



from __future__ import annotations

from contextlib import asynccontextmanager, contextmanager
from itertools import chain
from typing import TYPE_CHECKING, Any, Protocol, cast

from dishka import STRICT_VALIDATION, Scope, make_async_container
from dishka.async_container import CONTAINER_KEY
from dishka.entities.factory_type import FactoryType
from dishka.entities.marker import BaseMarker, BoolMarker

from waku.di import DEFAULT_COMPONENT, AsyncContainer, BaseProvider
from waku.extensions import DEFAULT_EXTENSIONS
from waku.factory import WakuFactory
from waku.modules import module

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Iterator, Sequence

    from dishka import Provider
    from dishka.dependency_source import Factory
    from dishka.registry import Registry

    from waku.application import WakuApplication
    from waku.extensions import ApplicationExtension, ModuleExtension
    from waku.modules import DynamicModule, ModuleType


__all__ = [
    'create_test_app',
    'override',
]


class _HasOverride(Protocol):
    when_override: BaseMarker | None


@contextmanager
def override(
    container: AsyncContainer,
    *providers: BaseProvider,
    context: dict[Any, Any] | None = None,
) -> Iterator[None]:
    """Temporarily override providers and/or context in an AsyncContainer for testing.

    Args:
        container: The container whose providers/context will be overridden.
        *providers: Providers to override in the container.
        context: Context values to override.

    Yields:
        None: Context in which the container uses the overridden providers/context.

    Example:
        ```python
        from waku import WakuFactory, module
        from waku.di import Scope, singleton
        from waku.testing import override


        class Service: ...


        class ServiceOverride(Service): ...


        # Override providers
        with override(application.container, singleton(ServiceOverride, provided_type=Service)):
            service = await application.container.get(Service)
            assert isinstance(service, ServiceOverride)

        # Override context
        with override(application.container, context={int: 123}):
            ...
        ```

    Raises:
        ValueError: If container is not at root (APP) scope.
    """
    if container.scope != Scope.APP:
        msg = (
            f'override() only supports root (APP scope) containers, '
            f'got {container.scope.name} scope. '
            f'Use application.container instead of a scoped container.'
        )
        raise ValueError(msg)

    _mark_as_overrides(providers)

    original_context = cast('dict[Any, Any]', container._context)  # noqa: SLF001
    merged_context = {**original_context, **(context or {})}
    new_container = make_async_container(
        _container_provider(container),
        *providers,
        context=merged_context,
        start_scope=container.scope,
        validation_settings=STRICT_VALIDATION,
    )

    # Only copy cache when no providers are overridden (context-only override)
    # Provider overrides may have transitive effects, so rebuild everything
    if not providers:
        _copy_cache(container, new_container)

    _swap(container, new_container)
    container._cache[CONTAINER_KEY] = container  # noqa: SLF001
    yield
    _swap(new_container, container)


def _container_provider(container: AsyncContainer) -> BaseProvider:
    container_provider = BaseProvider(component=DEFAULT_COMPONENT)
    container_provider.factories.extend(_extract_factories(container.registry))
    for registry in container.child_registries:
        container_provider.factories.extend(_extract_factories(registry))
    return container_provider


def _extract_factories(registry: Registry) -> list[Factory]:
    return [
        factory
        for dep_key, factory in registry.factories.items()
        if (dep_key.type_hint is not AsyncContainer and factory.type is not FactoryType.CONTEXT)
    ]


def _copy_cache(
    source: AsyncContainer,
    target: AsyncContainer,
) -> None:
    source_cache = cast('dict[Any, Any]', source._cache)  # noqa: SLF001
    target_cache = cast('dict[Any, Any]', target._cache)  # noqa: SLF001
    target_cache.update(source_cache)


def _swap(c1: AsyncContainer, c2: AsyncContainer) -> None:
    for attr in type(c1).__slots__:
        tmp = getattr(c1, attr)
        setattr(c1, attr, getattr(c2, attr))
        setattr(c2, attr, tmp)


@asynccontextmanager
async def create_test_app(
    *,
    base: ModuleType | DynamicModule | None = None,
    providers: Sequence[Provider] = (),
    imports: Sequence[ModuleType | DynamicModule] = (),
    extensions: Sequence[ModuleExtension] = (),
    app_extensions: Sequence[ApplicationExtension] = DEFAULT_EXTENSIONS,
    context: dict[Any, Any] | None = None,
) -> AsyncIterator[WakuApplication]:
    """Create a minimal test application with given configuration.

    Useful for testing extensions and module configurations in isolation
    without needing to set up a full application structure.

    Args:
        base: Base module to build upon. When provided, the test module
            imports this module and providers act as overrides.
        providers: Providers to register in the test module.
            When `base` is provided, these override existing providers.
        imports: Additional modules to import into the test module.
        extensions: Module extensions to register.
        app_extensions: Application extensions to register (default: DEFAULT_EXTENSIONS).
        context: Context values to pass to the container.

    Yields:
        Initialized WakuApplication.

    Example:
        ```python
        from waku.testing import create_test_app
        from waku.di import singleton


        class IRepository(Protocol):
            async def get(self, id: str) -> Entity: ...


        class FakeRepository(IRepository):
            async def get(self, id: str) -> Entity:
                return Entity(id=id)


        # Create test app from scratch
        async def test_my_extension():
            extension = MyExtension().bind(SomeEvent, SomeHandler)

            async with create_test_app(
                extensions=[extension],
                providers=[singleton(IRepository, FakeRepository)],
            ) as app:
                service = await app.container.get(MyService)
                result = await service.do_something()
                assert result == expected


        # Create test app based on existing module with overrides
        async def test_with_base_module():
            async with create_test_app(
                base=AppModule,
                providers=[singleton(IRepository, FakeRepository)],
            ) as app:
                # FakeRepository replaces the real one from AppModule
                repo = await app.container.get(IRepository)
                assert isinstance(repo, FakeRepository)
        ```
    """
    all_imports = list(imports)
    if base is not None:
        all_imports.insert(0, base)

    override_providers = list(providers)
    if base is not None:
        _mark_as_overrides(override_providers)

    @module(
        providers=override_providers,
        imports=all_imports,
        extensions=list(extensions),
    )
    class _TestModule:
        pass

    app = WakuFactory(_TestModule, context=context, extensions=app_extensions).create()
    async with app:
        yield app


def _mark_as_overrides(providers: Sequence[BaseProvider]) -> None:
    for prov in providers:
        for factory in chain[_HasOverride](prov.factories, prov.aliases):
            factory.when_override = BoolMarker(True)  # noqa: FBT003
        for context_var in prov.context_vars:
            context_var.override = True



from __future__ import annotations

import functools
import typing

from typing_extensions import get_original_bases

from waku.cqrs.contracts.request import IRequest

if typing.TYPE_CHECKING:
    from waku.cqrs.contracts import ResponseT

__all__ = ['get_request_response_type']


def _is_request_origin(origin: type | None) -> bool:
    if origin is None:
        return False
    if origin is IRequest:
        return True
    try:
        return isinstance(origin, type) and issubclass(origin, IRequest)  # pyrefly: ignore[invalid-argument]
    except TypeError:
        return False


def _extract_response_from_bases(cls: type) -> type[ResponseT] | None:
    for base in get_original_bases(cls):
        if not _is_request_origin(typing.get_origin(base)):
            continue
        if args := typing.get_args(base):
            response_type = args[0]
            if isinstance(response_type, typing.TypeVar):
                continue
            return typing.cast('type[ResponseT]', response_type)
    return None


@functools.cache
def get_request_response_type(request_type: type[IRequest[ResponseT]]) -> type[ResponseT]:
    """Extract the response type from an IRequest implementation.

    Searches through the class hierarchy to find IRequest or its subclasses,
    supporting direct implementations, Request subclasses, and nested inheritance.

    Raises:
        TypeError: if response type cannot be extracted from the request type.
    """
    for cls in request_type.__mro__:
        if cls is object:
            break
        if response_type := _extract_response_from_bases(cls):
            return response_type  # type: ignore[return-value]

    msg = f'Could not extract response type from {request_type.__name__}'
    raise TypeError(msg)



from __future__ import annotations

from sqlalchemy import BigInteger, Column, MetaData, Table, Text, func
from sqlalchemy.dialects.postgresql import TIMESTAMP

__all__ = ['bind_checkpoint_tables']

_internal_metadata = MetaData()

es_checkpoints_table = Table(
    'es_checkpoints',
    _internal_metadata,
    Column('projection_name', Text, primary_key=True),
    Column('position', BigInteger, nullable=False),
    Column('updated_at', TIMESTAMP(timezone=True), nullable=False),
    Column('created_at', TIMESTAMP(timezone=True), server_default=func.now()),
)


def bind_checkpoint_tables(metadata: MetaData) -> Table:
    return es_checkpoints_table.to_metadata(metadata)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator

    from waku.eventsourcing.projection.binding import CatchUpProjectionBinding

__all__ = ['CatchUpProjectionRegistry']


class CatchUpProjectionRegistry:
    def __init__(self, bindings: tuple[CatchUpProjectionBinding, ...]) -> None:
        self._bindings = bindings
        by_name: dict[str, CatchUpProjectionBinding] = {}
        for b in self._bindings:
            name = b.projection.projection_name
            if name in by_name:
                msg = f'Duplicate projection name {name!r}'
                raise ValueError(msg)
            by_name[name] = b
        self._by_name = by_name

    def __iter__(self) -> Iterator[CatchUpProjectionBinding]:
        return iter(self._bindings)

    def __len__(self) -> int:
        return len(self._bindings)

    def get(self, projection_name: str) -> CatchUpProjectionBinding:
        try:
            return self._by_name[projection_name]
        except KeyError:
            msg = f'Projection {projection_name!r} not found'
            raise ValueError(msg) from None



from __future__ import annotations

import abc
from typing import TYPE_CHECKING

from waku.eventsourcing.contracts.stream import StreamPosition

if TYPE_CHECKING:
    from collections.abc import Sequence

    from waku.eventsourcing.contracts.event import EventEnvelope, StoredEvent
    from waku.eventsourcing.contracts.stream import ExpectedVersion, StreamId

__all__ = [
    'IEventReader',
    'IEventStore',
    'IEventWriter',
]


class IEventReader(abc.ABC):
    @abc.abstractmethod
    async def read_stream(
        self,
        stream_id: StreamId,
        /,
        *,
        start: int | StreamPosition = StreamPosition.START,
        count: int | None = None,
    ) -> list[StoredEvent]: ...

    @abc.abstractmethod
    async def read_all(
        self,
        *,
        after_position: int = -1,
        count: int | None = None,
        event_types: Sequence[str] | None = None,
    ) -> list[StoredEvent]: ...

    @abc.abstractmethod
    async def stream_exists(self, stream_id: StreamId, /) -> bool: ...

    @abc.abstractmethod
    async def global_head_position(self) -> int:
        """Return the highest global position in the store, or ``-1`` if empty."""
        ...

    @abc.abstractmethod
    async def read_positions(
        self,
        *,
        after_position: int,
        up_to_position: int,
    ) -> list[int]:
        """Return committed global positions in the range ``(after_position, up_to_position]``.

        Positions are returned in ascending order.
        """
        ...


class IEventWriter(abc.ABC):
    @abc.abstractmethod
    async def append_to_stream(
        self,
        stream_id: StreamId,
        /,
        events: Sequence[EventEnvelope],
        *,
        expected_version: ExpectedVersion,
    ) -> int: ...


class IEventStore(IEventReader, IEventWriter, abc.ABC):
    pass



"""Extension protocols for application and module lifecycle hooks."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol, TypeAlias, runtime_checkable

if TYPE_CHECKING:
    from collections.abc import Mapping

    from waku.application import WakuApplication
    from waku.modules import Module, ModuleMetadata, ModuleMetadataRegistry, ModuleType

__all__ = [
    'AfterApplicationInit',
    'ApplicationExtension',
    'ModuleExtension',
    'OnApplicationInit',
    'OnApplicationShutdown',
    'OnModuleConfigure',
    'OnModuleDestroy',
    'OnModuleInit',
    'OnModuleRegistration',
]


@runtime_checkable
class OnApplicationInit(Protocol):
    """Extension for application pre-initialization actions."""

    __slots__ = ()

    async def on_app_init(self, app: WakuApplication) -> None: ...


@runtime_checkable
class AfterApplicationInit(Protocol):
    """Extension for application post-initialization actions."""

    __slots__ = ()

    async def after_app_init(self, app: WakuApplication) -> None: ...


@runtime_checkable
class OnApplicationShutdown(Protocol):
    """Extension for application shutdown actions."""

    __slots__ = ()

    async def on_app_shutdown(self, app: WakuApplication) -> None: ...


@runtime_checkable
class OnModuleRegistration(Protocol):
    """Extension for contributing providers to module metadata during registration.

    This hook runs after all module metadata is collected but before Module
    objects are created. Use this for cross-module aggregation that produces
    providers which should belong to the owning module.

    Can be declared at both application level (passed to WakuFactory) and
    module level (in module's extensions list).

    Execution order:
        1. Application-level extensions (assigned to root module)
        2. Module-level extensions (in topological order)

    Key differences from OnModuleConfigure:
        - Runs after ALL modules are collected (cross-module visibility)
        - Receives registry with access to all modules' metadata
        - Can add providers to owning module
    """

    __slots__ = ()

    def on_module_registration(
        self,
        registry: ModuleMetadataRegistry,
        owning_module: ModuleType,
        context: Mapping[Any, Any] | None,
    ) -> None:
        """Contribute providers to module metadata before Module objects are created.

        Args:
            registry: Registry of all collected module metadata. Use find_extensions()
                      to discover extensions across modules, add_provider() to contribute.
            owning_module: The module type that owns this extension. Providers
                          added via registry.add_provider() should target this module.
            context: Application context passed to WakuFactory (read-only).
        """
        ...


@runtime_checkable
class OnModuleConfigure(Protocol):
    """Extension for module configuration."""

    __slots__ = ()

    def on_module_configure(self, metadata: ModuleMetadata) -> None:
        """Perform actions before module metadata transformed to module."""
        ...


@runtime_checkable
class OnModuleInit(Protocol):
    """Extension for module initialization."""

    __slots__ = ()

    async def on_module_init(self, module: Module) -> None: ...


@runtime_checkable
class OnModuleDestroy(Protocol):
    """Extension for module destroying."""

    __slots__ = ()

    async def on_module_destroy(self, module: Module) -> None: ...


ApplicationExtension: TypeAlias = (
    OnApplicationInit | AfterApplicationInit | OnApplicationShutdown | OnModuleRegistration
)
ModuleExtension: TypeAlias = OnModuleConfigure | OnModuleInit | OnModuleDestroy | OnModuleRegistration



from __future__ import annotations

from abc import ABC, abstractmethod
from itertools import chain
from typing import TYPE_CHECKING

from dishka.entities.factory_type import FactoryType
from typing_extensions import override

from waku.di import Scope
from waku.validation import ValidationError, ValidationRule
from waku.validation.rules._cache import LRUCache
from waku.validation.rules._types_extractor import ModuleTypesExtractor

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from dishka import AsyncContainer
    from dishka.entities.key import DependencyKey

    from waku.modules import Module, ModuleRegistry
    from waku.validation._extension import ValidationContext


__all__ = [
    'DependenciesAccessibleRule',
    'DependencyInaccessibleError',
]


class DependencyInaccessibleError(ValidationError):
    """Error indicating a dependency is not accessible to a provider/module."""

    def __init__(
        self,
        required_type: type[object],
        required_by: object,
        from_module: Module,
    ) -> None:
        self.required_type = required_type
        self.required_by = required_by
        self.from_module = from_module
        super().__init__(str(self))

    def __str__(self) -> str:
        msg = [
            f'Dependency Error: "{self.required_type!r}" is not accessible',
            f'Required by: "{self.required_by!r}"',
            f'In module: "{self.from_module!r}"',
            '',
            'To resolve this issue, either:',
            f'1. Export "{self.required_type!r}" from a module that provides it and add that module to "{self.from_module!r}" imports',
            f'2. Make the module that provides "{self.required_type!r}" global by setting is_global=True',
            f'3. Move the dependency to a module that has access to "{self.required_type!r}"',
            '',
            'Note: Dependencies can only be accessed from:',
            '- The same module that provides them',
            '- Modules that import the module that provides and exports it',
            '- Global modules',
        ]
        return '\n'.join(msg)


class AccessibilityStrategy(ABC):
    """Strategy for checking if a type is accessible to a module."""

    __slots__ = ()

    @abstractmethod
    def is_accessible(self, required_type: type[object], module: Module) -> bool:
        """Check if the required type is accessible to the given module."""


class GlobalProvidersStrategy(AccessibilityStrategy):
    """Check if type is provided by a global module or APP-scoped context."""

    __slots__ = ('_global_types',)

    def __init__(
        self,
        modules: Sequence[Module],
        container: AsyncContainer,
        types_extractor: ModuleTypesExtractor,
        registry: ModuleRegistry,
    ) -> None:
        self._global_types = self._build_global_types(modules, container, types_extractor, registry)

    @override
    def is_accessible(self, required_type: type[object], module: Module) -> bool:
        return required_type in self._global_types

    @staticmethod
    def _build_global_types(
        modules: Sequence[Module],
        container: AsyncContainer,
        types_extractor: ModuleTypesExtractor,
        registry: ModuleRegistry,
    ) -> frozenset[type[object]]:
        global_module_types = {
            provided_type
            for mod in modules
            if mod.is_global
            for provided_type in chain(
                types_extractor.get_provided_types(mod),
                types_extractor.get_reexported_types(mod, registry),
            )
        }

        global_context_types = {
            dep.type_hint
            for dep, factory in container.registry.factories.items()
            if factory.scope is Scope.APP and factory.type is FactoryType.CONTEXT
        }

        return frozenset(global_module_types | global_context_types)


class LocalProvidersStrategy(AccessibilityStrategy):
    """Check if type is provided by the module itself."""

    __slots__ = ('_types_extractor',)

    def __init__(self, types_extractor: ModuleTypesExtractor) -> None:
        self._types_extractor = types_extractor

    @override
    def is_accessible(self, required_type: type[object], module: Module) -> bool:
        return required_type in self._types_extractor.get_provided_types(module)


class ContextVarsStrategy(AccessibilityStrategy):
    """Check if type is provided by application or request container context."""

    __slots__ = ('_types_extractor',)

    def __init__(self, types_extractor: ModuleTypesExtractor) -> None:
        self._types_extractor = types_extractor

    @override
    def is_accessible(self, required_type: type[object], module: Module) -> bool:
        return required_type in self._types_extractor.get_context_vars(module)


class ImportedModulesStrategy(AccessibilityStrategy):
    """Check if type is accessible via imported modules (direct export or re-export)."""

    __slots__ = ('_registry', '_types_extractor')

    def __init__(self, registry: ModuleRegistry, types_extractor: ModuleTypesExtractor) -> None:
        self._registry = registry
        self._types_extractor = types_extractor

    @override
    def is_accessible(self, required_type: type[object], module: Module) -> bool:
        for imported in module.imports:
            imported_module = self._registry.get(imported)
            if self._is_directly_exported(required_type, imported_module):
                return True
            if self._is_reexported(required_type, imported_module):
                return True
        return False

    def _is_directly_exported(self, required_type: type[object], imported_module: Module) -> bool:
        return (
            required_type in self._types_extractor.get_provided_types(imported_module)
            and required_type in imported_module.exports
        )

    def _is_reexported(self, required_type: type[object], imported_module: Module) -> bool:
        return required_type in self._types_extractor.get_reexported_types(imported_module, self._registry)


class DependencyAccessChecker:
    """Handles dependency accessibility checks between modules."""

    __slots__ = ('_strategies',)

    def __init__(self, strategies: Sequence[AccessibilityStrategy]) -> None:
        self._strategies = strategies

    def find_inaccessible_dependencies(
        self,
        dependencies: Sequence[DependencyKey],
        module: Module,
    ) -> Iterable[type[object]]:
        for dependency in dependencies:
            if not self._is_accessible(dependency.type_hint, module):
                yield dependency.type_hint

    def _is_accessible(self, required_type: type[object], module: Module) -> bool:
        return any(strategy.is_accessible(required_type, module) for strategy in self._strategies)


class DependenciesAccessibleRule(ValidationRule):
    """Validates that all dependencies required by providers are accessible."""

    __slots__ = ('_cache', '_types_extractor')

    def __init__(self, cache_size: int = 1000) -> None:
        self._cache = LRUCache[set[type[object]]](cache_size)
        self._types_extractor = ModuleTypesExtractor(self._cache)

    @override
    def validate(self, context: ValidationContext) -> list[ValidationError]:
        self._cache.clear()

        registry = context.app.registry
        modules = list(registry.modules)
        container = context.app.container

        strategies: list[AccessibilityStrategy] = [
            GlobalProvidersStrategy(modules, container, self._types_extractor, registry),
            LocalProvidersStrategy(self._types_extractor),
            ContextVarsStrategy(self._types_extractor),
            ImportedModulesStrategy(registry, self._types_extractor),
        ]

        checker = DependencyAccessChecker(strategies)
        errors: list[ValidationError] = []

        for module in modules:
            for factory in module.provider.factories:
                inaccessible_deps = checker.find_inaccessible_dependencies(
                    dependencies=factory.dependencies,
                    module=module,
                )
                errors.extend(
                    DependencyInaccessibleError(
                        required_type=dep_type,
                        required_by=factory.source,
                        from_module=module,
                    )
                    for dep_type in inaccessible_deps
                )

        return errors



from waku.validation.rules.dependency_accessible import (
    DependenciesAccessibleRule,
    DependencyInaccessibleError,
)

__all__ = [
    'DependenciesAccessibleRule',
    'DependencyInaccessibleError',
]



from collections.abc import Sequence
from typing import Any, cast, overload

from dishka.exceptions import NoFactoryError
from typing_extensions import override

from waku.cqrs import IPipelineBehavior
from waku.cqrs.contracts.notification import INotification, NotificationT
from waku.cqrs.contracts.request import IRequest, ResponseT
from waku.cqrs.events.handler import INotificationHandler
from waku.cqrs.events.publish import EventPublisher
from waku.cqrs.exceptions import RequestHandlerNotFound
from waku.cqrs.interfaces import IMediator
from waku.cqrs.pipeline import PipelineBehaviorWrapper
from waku.cqrs.registry import MediatorRegistry
from waku.cqrs.requests.handler import RequestHandler
from waku.di import AsyncContainer


class Mediator(IMediator):
    """Default CQRS implementation."""

    __slots__ = ('_container', '_event_publisher', '_registry')

    def __init__(
        self,
        container: AsyncContainer,
        event_publisher: EventPublisher,
        registry: MediatorRegistry,
    ) -> None:
        self._container = container
        self._event_publisher = event_publisher
        self._registry = registry

    @overload
    async def send(self, request: IRequest[None], /) -> None: ...

    @overload
    async def send(self, request: IRequest[ResponseT], /) -> ResponseT: ...

    @override
    async def send(self, request: IRequest[Any], /) -> Any:
        request_type = type(request)
        handler = await self._resolve_request_handler(request_type)  # pyrefly: ignore[bad-argument-type]
        return await self._handle_request(handler, request)

    @override
    async def publish(self, notification: INotification, /) -> None:
        event_type = type(notification)
        handlers = await self._resolve_event_handlers(event_type)
        await self._event_publisher(handlers, notification)

    async def _resolve_request_handler(
        self,
        request_type: type[IRequest[ResponseT]],
    ) -> RequestHandler[IRequest[ResponseT], ResponseT]:
        if not self._registry.request_map.has_handler(request_type):
            raise RequestHandlerNotFound(request_type)

        handler_type = self._registry.request_map.get_handler_type(request_type)
        return cast('RequestHandler[IRequest[ResponseT], ResponseT]', await self._container.get(handler_type))

    async def _handle_request(
        self,
        handler: RequestHandler[IRequest[ResponseT], ResponseT],
        request: IRequest[ResponseT],
    ) -> ResponseT:
        request_type = type(request)
        behaviors = await self._resolve_behaviors(request_type)  # pyrefly: ignore[bad-argument-type]

        pipeline = PipelineBehaviorWrapper(behaviors).wrap(handler.handle)
        result = await pipeline(request)

        return cast('ResponseT', result)

    async def _resolve_behaviors(self, request_type: type[IRequest[Any]]) -> Sequence[IPipelineBehavior[Any, Any]]:
        try:
            global_behaviors = await self._container.get(Sequence[IPipelineBehavior[Any, Any]])
        except NoFactoryError:
            global_behaviors = []

        request_specific_behaviors: Sequence[IPipelineBehavior[Any, Any]] = []
        if self._registry.behavior_map.has_behaviors(request_type):
            lookup_type = self._registry.behavior_map.get_lookup_type(request_type)
            request_specific_behaviors = await self._container.get(Sequence[lookup_type])  # type: ignore[valid-type]

        return [*global_behaviors, *request_specific_behaviors]

    async def _resolve_event_handlers(
        self,
        event_type: type[NotificationT],
    ) -> Sequence[INotificationHandler[NotificationT]]:
        if not self._registry.event_map.has_handlers(event_type):
            return []

        handler_type = self._registry.event_map.get_handler_type(event_type)
        handlers = await self._container.get(Sequence[handler_type])  # type: ignore[valid-type]
        return cast('Sequence[INotificationHandler[NotificationT]]', handlers)



from waku.cqrs.events.handler import EventHandler
from waku.cqrs.events.publish import EventPublisher, GroupEventPublisher, SequentialEventPublisher

__all__ = [
    'EventHandler',
    'EventPublisher',
    'GroupEventPublisher',
    'SequentialEventPublisher',
]



from __future__ import annotations

from typing import TYPE_CHECKING, Final, cast

from waku.di import DEFAULT_COMPONENT, BaseProvider

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence
    from uuid import UUID

    from dishka import Provider

    from waku.extensions import ModuleExtension
    from waku.modules._metadata import DynamicModule, ModuleMetadata, ModuleType


__all__ = ['Module']


class Module:
    __slots__ = (
        '_provider',
        'exports',
        'extensions',
        'id',
        'imports',
        'is_global',
        'providers',
        'target',
    )

    def __init__(self, module_type: ModuleType, metadata: ModuleMetadata) -> None:
        self.id: Final[UUID] = metadata.id
        self.target: Final[ModuleType] = module_type

        self.providers: Final[Sequence[Provider]] = metadata.providers
        self.imports: Final[Sequence[ModuleType | DynamicModule]] = metadata.imports
        self.exports: Final[Sequence[type[object] | ModuleType | DynamicModule]] = metadata.exports
        self.extensions: Final[Sequence[ModuleExtension]] = metadata.extensions
        self.is_global: Final[bool] = metadata.is_global

        self._provider: BaseProvider | None = None

    @property
    def name(self) -> str:
        return self.target.__name__

    @property
    def provider(self) -> BaseProvider:
        if self._provider is None:
            msg = f'Module {self.name} provider not yet created. Call create_provider() first.'
            raise RuntimeError(msg)
        return self._provider

    def create_provider(self) -> BaseProvider:
        cls = cast('type[_ModuleProvider]', type(f'{self.name}Provider', (_ModuleProvider,), {}))
        self._provider = cls(self.providers)
        return self._provider

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f'Module[{self.name}]'

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        return self.id == other.id if isinstance(other, Module) else False


class _ModuleProvider(BaseProvider):
    def __init__(self, providers: Iterable[Provider]) -> None:
        super().__init__(DEFAULT_COMPONENT)
        for provider in providers:
            self.factories.extend(provider.factories)
            self.aliases.extend(provider.aliases)
            self.decorators.extend(provider.decorators)
            self.context_vars.extend(provider.context_vars)
            self.factory_union_mode.extend(provider.factory_union_mode)
            self.activators.extend(provider.activators)



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Protocol, TypeVar

from waku.cqrs.contracts.notification import INotification

if TYPE_CHECKING:
    from collections.abc import Sequence

__all__ = [
    'CommandT',
    'EventSourcedAggregate',
    'EventT',
    'IDecider',
    'StateT',
]

StateT = TypeVar('StateT')
CommandT = TypeVar('CommandT', contravariant=True)  # noqa: PLC0105
EventT = TypeVar('EventT', bound=INotification)


class IDecider(Protocol[StateT, CommandT, EventT]):
    def initial_state(self) -> StateT: ...
    def decide(self, command: CommandT, state: StateT) -> Sequence[EventT]: ...
    def evolve(self, state: StateT, event: EventT) -> StateT: ...


class EventSourcedAggregate(abc.ABC):
    _version: int
    _pending_events: list[INotification]

    def __init__(self) -> None:
        self._version = -1
        self._pending_events = []

    @property
    def version(self) -> int:
        return self._version

    def collect_events(self) -> list[INotification]:
        events = list(self._pending_events)
        self._pending_events.clear()
        return events

    def mark_persisted(self, version: int) -> None:
        self._version = version

    def _raise_event(self, event: INotification) -> None:
        self._apply(event)
        self._pending_events.append(event)

    @abc.abstractmethod
    def _apply(self, event: INotification) -> None: ...

    def load_from_history(self, events: Sequence[INotification], version: int) -> None:
        for event in events:
            self._apply(event)
        self._version = version



from __future__ import annotations

from collections.abc import MutableMapping
from dataclasses import dataclass
from typing import Generic, Self, TypeAlias

from typing_extensions import TypeVar

from waku.cqrs.contracts.request import IRequest, RequestT, Response, ResponseT
from waku.cqrs.exceptions import MapFrozenError, RequestHandlerAlreadyRegistered
from waku.cqrs.requests.handler import RequestHandler
from waku.cqrs.utils import get_request_response_type

__all__ = [
    'RequestMap',
    'RequestMapEntry',
    'RequestMapRegistry',
]

_MapReqT = TypeVar('_MapReqT', bound='IRequest[Response | None]', default='IRequest[Response | None]')
_MapResT = TypeVar('_MapResT', bound='Response | None', default='Response | None')


@dataclass(frozen=True, slots=True)
class RequestMapEntry(Generic[_MapReqT, _MapResT]):
    handler_type: type[RequestHandler[_MapReqT, _MapResT]]
    di_lookup_type: type[RequestHandler[_MapReqT, _MapResT]]


RequestMapRegistry: TypeAlias = MutableMapping[
    type[IRequest[Response | None]], RequestMapEntry[IRequest[Response | None], Response | None]
]


class RequestMap:
    def __init__(self) -> None:
        self._registry: RequestMapRegistry = {}
        self._frozen = False

    def freeze(self) -> None:
        self._frozen = True

    @property
    def is_frozen(self) -> bool:
        return self._frozen

    def bind(
        self,
        request_type: type[RequestT],
        handler_type: type[RequestHandler[RequestT, ResponseT]],
    ) -> Self:
        if self._frozen:
            raise MapFrozenError
        if request_type in self._registry:
            raise RequestHandlerAlreadyRegistered(request_type, handler_type)
        response_type = get_request_response_type(request_type)
        di_lookup_type = RequestHandler[request_type, response_type]  # type: ignore[valid-type]
        self._registry[request_type] = RequestMapEntry(handler_type, di_lookup_type)  # type: ignore[type-abstract, arg-type]
        return self

    def merge(self, other: RequestMap) -> Self:
        if self._frozen:
            raise MapFrozenError
        for request_type, entry in other._registry.items():
            self.bind(request_type, entry.handler_type)  # ty: ignore[invalid-argument-type]
        return self

    @property
    def registry(self) -> RequestMapRegistry:
        return self._registry

    def has_handler(self, request_type: type[RequestT]) -> bool:
        return request_type in self._registry

    def get_handler_type(self, request_type: type[RequestT]) -> type[RequestHandler[RequestT, Response | None]]:
        return self._registry[request_type].di_lookup_type

    def __bool__(self) -> bool:
        return bool(self._registry)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, TypeVar

from waku.eventsourcing.exceptions import ConcurrencyConflictError

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

logger = logging.getLogger(__name__)

_T = TypeVar('_T')


async def execute_with_optimistic_retry(
    attempt_fn: Callable[[], Awaitable[_T]],
    *,
    max_attempts: int,
    is_creation: bool,
    request_name: str,
    aggregate_id: str,
) -> _T:
    for attempt in range(1, max_attempts + 1):
        if attempt > 1:
            logger.info(
                'Retrying %s for %s (attempt %d/%d) after concurrency conflict',
                request_name,
                aggregate_id,
                attempt,
                max_attempts,
            )

        try:
            return await attempt_fn()
        except ConcurrencyConflictError:
            if is_creation or attempt == max_attempts:
                raise

    msg = 'max_attempts must be >= 1'  # pragma: no cover
    raise ValueError(msg)  # pragma: no cover



from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from waku.eventsourcing.contracts.stream import StreamId

__all__ = [
    'ISnapshotStore',
    'ISnapshotStrategy',
    'Snapshot',
]


@dataclass(frozen=True, slots=True, kw_only=True)
class Snapshot:
    stream_id: StreamId
    state: dict[str, Any]
    version: int
    state_type: str
    schema_version: int = 1


class ISnapshotStore(abc.ABC):
    @abc.abstractmethod
    async def load(self, stream_id: StreamId, /) -> Snapshot | None: ...

    @abc.abstractmethod
    async def save(self, snapshot: Snapshot, /) -> None: ...


class ISnapshotStrategy(abc.ABC):
    @abc.abstractmethod
    def should_snapshot(self, version: int, events_since_snapshot: int) -> bool: ...



"""Extension registry for centralized management of extensions."""

from __future__ import annotations

import inspect
from collections import defaultdict
from typing import TYPE_CHECKING, Self, TypeVar, cast

from waku.extensions.protocols import ApplicationExtension, ModuleExtension

if TYPE_CHECKING:
    from waku.modules import ModuleType

__all__ = ['ExtensionRegistry']


_AppExtT = TypeVar('_AppExtT', bound=ApplicationExtension)
_ModExtT = TypeVar('_ModExtT', bound=ModuleExtension)


class ExtensionRegistry:
    """Registry for extensions.

    This registry maintains references to all extensions in the application,
    allowing for centralized management and discovery.
    """

    def __init__(self) -> None:
        self._app_extensions: dict[type[ApplicationExtension], list[ApplicationExtension]] = defaultdict(list)
        self._module_extensions: dict[ModuleType, list[ModuleExtension]] = defaultdict(list)

    def register_application_extension(self, extension: ApplicationExtension) -> Self:
        """Register an application extension with optional priority and tags."""
        ext_type = type(extension)
        extension_bases = [
            base
            for base in inspect.getmro(ext_type)
            if (isinstance(base, ApplicationExtension) and base != ext_type)  # type: ignore[unreachable]
        ]
        for base in extension_bases:
            self._app_extensions[cast('type[ApplicationExtension]', base)].append(extension)
        return self

    def register_module_extension(self, module_type: ModuleType, extension: ModuleExtension) -> Self:
        self._module_extensions[module_type].append(extension)
        return self

    def get_application_extensions(self, extension_type: type[_AppExtT]) -> list[_AppExtT]:
        return cast('list[_AppExtT]', self._app_extensions.get(cast('type[ApplicationExtension]', extension_type), []))

    def get_module_extensions(self, module_type: ModuleType, extension_type: type[_ModExtT]) -> list[_ModExtT]:
        extensions = cast('list[_ModExtT]', self._module_extensions.get(module_type, []))
        return [ext for ext in extensions if isinstance(ext, extension_type)]



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field, replace
from typing import TYPE_CHECKING, Any, Self, TypeAlias

from typing_extensions import override

from waku.di import Provider, WithParents, many, object_, scoped
from waku.eventsourcing._introspection import resolve_generic_args
from waku.eventsourcing.contracts.aggregate import IDecider
from waku.eventsourcing.contracts.event import IMetadataEnricher
from waku.eventsourcing.decider.repository import DeciderRepository
from waku.eventsourcing.exceptions import (
    DuplicateAggregateNameError,
    EventSourcingConfigError,
    RegistryFrozenError,
    SnapshotMigrationChainError,
    UnknownEventTypeError,
    UpcasterChainError,
)
from waku.eventsourcing.projection.binding import CatchUpProjectionBinding
from waku.eventsourcing.projection.interfaces import ErrorPolicy, ICatchUpProjection, ICheckpointStore, IProjection
from waku.eventsourcing.projection.registry import CatchUpProjectionRegistry
from waku.eventsourcing.serialization.interfaces import IEventSerializer, ISnapshotStateSerializer
from waku.eventsourcing.serialization.registry import EventTypeRegistry
from waku.eventsourcing.snapshot.interfaces import ISnapshotStore
from waku.eventsourcing.snapshot.migration import SnapshotMigrationChain
from waku.eventsourcing.snapshot.registry import SnapshotConfig, SnapshotConfigRegistry
from waku.eventsourcing.store.interfaces import IEventStore
from waku.eventsourcing.upcasting.chain import UpcasterChain
from waku.extensions import OnModuleConfigure, OnModuleRegistration
from waku.modules import DynamicModule, ModuleMetadataRegistry, module

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Mapping, Sequence

    from waku.cqrs.contracts.notification import INotification
    from waku.eventsourcing.repository import EventSourcedRepository
    from waku.eventsourcing.snapshot.interfaces import ISnapshotStrategy
    from waku.eventsourcing.snapshot.migration import ISnapshotMigration
    from waku.eventsourcing.upcasting.interfaces import IEventUpcaster
    from waku.modules import ModuleMetadata, ModuleType


__all__ = [
    'EventSourcingConfig',
    'EventSourcingExtension',
    'EventSourcingModule',
    'EventType',
    'EventTypeSpec',
    'SnapshotOptions',
]


@dataclass(frozen=True, slots=True)
class EventType:
    event_type: type[INotification]
    name: str | None = field(default=None, kw_only=True)
    aliases: Sequence[str] = field(default=(), kw_only=True)
    version: int = field(default=1, kw_only=True)
    upcasters: Sequence[IEventUpcaster] = field(default=(), kw_only=True)


EventTypeSpec: TypeAlias = 'type[INotification] | EventType'


@dataclass(frozen=True, slots=True, kw_only=True)
class SnapshotOptions:
    strategy: ISnapshotStrategy
    schema_version: int = 1
    migrations: Sequence[ISnapshotMigration] = ()


@dataclass(frozen=True, slots=True, kw_only=True)
class AggregateBinding:
    repository: type[EventSourcedRepository[Any]]
    event_types: Sequence[EventTypeSpec]
    projections: Sequence[type[IProjection]]
    snapshot: SnapshotOptions | None


@dataclass(frozen=True, slots=True, kw_only=True)
class DeciderBinding:
    repository: type[DeciderRepository[Any, Any, Any]]
    decider: type[IDecider[Any, Any, Any]]
    event_types: Sequence[EventTypeSpec]
    projections: Sequence[type[IProjection]]
    snapshot: SnapshotOptions | None


@dataclass(frozen=True, slots=True, kw_only=True)
class EventSourcingConfig:
    store: type[IEventStore] | Callable[..., IEventStore]
    event_serializer: type[IEventSerializer] | Callable[..., IEventSerializer] | None = None
    snapshot_store: type[ISnapshotStore] | Callable[..., ISnapshotStore] | None = None
    snapshot_state_serializer: type[ISnapshotStateSerializer] | Callable[..., ISnapshotStateSerializer] | None = None
    checkpoint_store: type[ICheckpointStore] | Callable[..., ICheckpointStore] | None = None
    enrichers: Sequence[type[IMetadataEnricher]] = ()


@dataclass(slots=True)
class EventSourcingRegistry:
    projection_types: list[type[IProjection]] = field(default_factory=list)
    catch_up_projection_types: list[type[ICatchUpProjection]] = field(default_factory=list)
    event_type_bindings: list[EventTypeSpec] = field(default_factory=list)
    _frozen: bool = field(default=False, init=False, repr=False)

    def merge(self, other: EventSourcingRegistry) -> None:
        self._check_not_frozen()
        self.projection_types.extend(other.projection_types)
        self.catch_up_projection_types.extend(other.catch_up_projection_types)
        self.event_type_bindings.extend(other.event_type_bindings)

    def freeze(self) -> None:
        self._frozen = True

    def handler_providers(self) -> Iterator[Provider]:
        if self.projection_types:
            yield many(IProjection, *self.projection_types, collect=False)
        if self.catch_up_projection_types:
            yield many(ICatchUpProjection, *self.catch_up_projection_types, collect=False)

    @staticmethod
    def collector_providers() -> Iterator[Provider]:
        yield many(IProjection, collect=True)
        yield many(ICatchUpProjection, collect=True)

    def _check_not_frozen(self) -> None:
        if self._frozen:
            raise RegistryFrozenError


@module()
class EventSourcingModule:
    @classmethod
    def register(cls, config: EventSourcingConfig, /) -> DynamicModule:
        providers: list[Provider] = [
            scoped(IEventStore, config.store),
        ]

        if config.event_serializer is not None:
            providers.append(scoped(IEventSerializer, config.event_serializer))

        if config.snapshot_store is not None:
            providers.append(scoped(ISnapshotStore, config.snapshot_store))

        if config.snapshot_state_serializer is not None:
            providers.append(scoped(ISnapshotStateSerializer, config.snapshot_state_serializer))

        if config.checkpoint_store is not None:
            providers.append(scoped(ICheckpointStore, config.checkpoint_store))

        providers.append(many(IMetadataEnricher, *config.enrichers))

        return DynamicModule(
            parent_module=cls,
            providers=providers,
            extensions=[
                EventSourcingRegistryAggregator(has_serializer=config.event_serializer is not None),
            ],
            is_global=True,
        )


@dataclass
class EventSourcingExtension(OnModuleConfigure):
    _bindings: list[AggregateBinding] = field(default_factory=list, init=False)
    _decider_bindings: list[DeciderBinding] = field(default_factory=list, init=False)
    _catch_up_bindings: list[CatchUpProjectionBinding] = field(default_factory=list, init=False)
    _registry: EventSourcingRegistry = field(default_factory=EventSourcingRegistry, init=False)

    def bind_aggregate(
        self,
        repository: type[EventSourcedRepository[Any]],
        event_types: Sequence[EventTypeSpec] = (),
        projections: Sequence[type[IProjection]] = (),
        snapshot: SnapshotOptions | None = None,
    ) -> Self:
        self._bindings.append(
            AggregateBinding(
                repository=repository,
                event_types=event_types,
                projections=projections,
                snapshot=snapshot,
            )
        )
        self._registry.projection_types.extend(projections)
        self._registry.event_type_bindings.extend(event_types)
        return self

    def bind_decider(
        self,
        repository: type[DeciderRepository[Any, Any, Any]],
        decider: type[IDecider[Any, Any, Any]],
        event_types: Sequence[EventTypeSpec] = (),
        projections: Sequence[type[IProjection]] = (),
        snapshot: SnapshotOptions | None = None,
    ) -> Self:
        self._decider_bindings.append(
            DeciderBinding(
                repository=repository,
                decider=decider,
                event_types=event_types,
                projections=projections,
                snapshot=snapshot,
            )
        )
        self._registry.projection_types.extend(projections)
        self._registry.event_type_bindings.extend(event_types)
        return self

    def bind_catch_up_projection(  # noqa: PLR0913
        self,
        projection: type[ICatchUpProjection],
        *,
        error_policy: ErrorPolicy = ErrorPolicy.STOP,
        max_retry_attempts: int = 0,
        base_retry_delay_seconds: float = 10.0,
        max_retry_delay_seconds: float = 300.0,
        batch_size: int = 100,
        gap_detection_enabled: bool = False,
        gap_timeout_seconds: float = 10.0,
    ) -> Self:
        self._registry.catch_up_projection_types.append(projection)
        self._catch_up_bindings.append(
            CatchUpProjectionBinding(
                projection=projection,
                error_policy=error_policy,
                max_retry_attempts=max_retry_attempts,
                base_retry_delay_seconds=base_retry_delay_seconds,
                max_retry_delay_seconds=max_retry_delay_seconds,
                batch_size=batch_size,
                gap_detection_enabled=gap_detection_enabled,
                gap_timeout_seconds=gap_timeout_seconds,
            )
        )
        return self

    @property
    def catch_up_bindings(self) -> list[CatchUpProjectionBinding]:
        return self._catch_up_bindings

    @property
    def registry(self) -> EventSourcingRegistry:
        return self._registry

    def aggregate_names(self) -> Iterator[tuple[str, type]]:
        for binding in self._bindings:
            yield binding.repository.aggregate_name, binding.repository
        for binding in self._decider_bindings:
            yield binding.repository.aggregate_name, binding.repository

    def snapshot_bindings(self) -> Iterator[tuple[str, SnapshotOptions]]:
        for binding in self._bindings:
            if binding.snapshot is not None:
                yield binding.repository.aggregate_name, binding.snapshot
        for binding in self._decider_bindings:
            if binding.snapshot is not None:
                yield binding.repository.aggregate_name, binding.snapshot

    def on_module_configure(self, metadata: ModuleMetadata) -> None:
        for binding in self._bindings:
            repo_type = binding.repository
            metadata.providers.append(scoped(WithParents[repo_type], repo_type))  # type: ignore[misc,valid-type]

        for binding in self._decider_bindings:
            repo_type = binding.repository
            metadata.providers.append(scoped(WithParents[repo_type], repo_type))  # type: ignore[misc,valid-type]
            decider_iface = self._resolve_decider_interface(repo_type)
            metadata.providers.append(scoped(decider_iface, binding.decider))

    @staticmethod
    def _resolve_decider_interface(repo_type: type[Any]) -> type[Any]:
        args = resolve_generic_args(repo_type, DeciderRepository)
        if args and len(args) == 3:  # noqa: PLR2004
            return IDecider[args[0], args[1], args[2]]  # type: ignore[valid-type]
        return IDecider  # pragma: no cover


class EventSourcingRegistryAggregator(OnModuleRegistration):
    def __init__(self, *, has_serializer: bool = False) -> None:
        self._has_serializer = has_serializer

    @override
    def on_module_registration(
        self,
        registry: ModuleMetadataRegistry,
        owning_module: ModuleType,
        context: Mapping[Any, Any] | None,
    ) -> None:
        aggregated = EventSourcingRegistry()
        all_aggregate_names: defaultdict[str, list[type]] = defaultdict(list)
        all_catch_up_bindings: list[CatchUpProjectionBinding] = []

        for module_type, ext in registry.find_extensions(EventSourcingExtension):
            aggregated.merge(ext.registry)
            all_catch_up_bindings.extend(ext.catch_up_bindings)
            for provider in ext.registry.handler_providers():
                registry.add_provider(module_type, provider)
            for name, repo_type in ext.aggregate_names():
                all_aggregate_names[name].append(repo_type)

        for name, repos in all_aggregate_names.items():
            if len(repos) > 1:
                raise DuplicateAggregateNameError(name, repos)

        for provider in aggregated.collector_providers():
            registry.add_provider(owning_module, provider)

        event_type_registry, upcaster_chain = self._build_type_registry(aggregated)
        registry.add_provider(owning_module, object_(event_type_registry))
        registry.add_provider(owning_module, object_(upcaster_chain))

        resolved_bindings = self._resolve_catch_up_bindings(all_catch_up_bindings, event_type_registry)
        registry.add_provider(
            owning_module,
            object_(CatchUpProjectionRegistry(resolved_bindings)),
        )

        aggregated.freeze()
        registry.add_provider(owning_module, object_(aggregated))

        snapshot_config_registry = self._build_snapshot_config_registry(
            registry.find_extensions(EventSourcingExtension),
        )
        registry.add_provider(owning_module, object_(snapshot_config_registry))

        if self._has_serializer and len(event_type_registry) == 0:
            msg = (
                'A serializer is configured but no event types were registered via '
                'bind_aggregate(event_types=[...]). Deserialization will fail at runtime.'
            )
            raise EventSourcingConfigError(msg)

    @staticmethod
    def _resolve_catch_up_bindings(
        bindings: Sequence[CatchUpProjectionBinding],
        event_type_registry: EventTypeRegistry,
    ) -> tuple[CatchUpProjectionBinding, ...]:
        resolved: list[CatchUpProjectionBinding] = []
        for binding in bindings:
            if binding.projection.event_types is None:
                resolved.append(binding)
                continue
            names: list[str] = []
            for et in binding.projection.event_types:
                try:
                    names.append(event_type_registry.get_name(et))
                except UnknownEventTypeError:
                    msg = (
                        f'Projection {binding.projection.__name__!r} declares event type '
                        f'{et.__name__!r} in its event_types, but it is not registered '
                        f'via bind_aggregate(event_types=[...]).'
                    )
                    raise EventSourcingConfigError(msg) from None
            resolved.append(replace(binding, event_type_names=tuple(names)))
        return tuple(resolved)

    @staticmethod
    def _build_snapshot_config_registry(
        extensions: Iterator[tuple[ModuleType, EventSourcingExtension]],
    ) -> SnapshotConfigRegistry:
        configs: dict[str, SnapshotConfig] = {}
        for _module_type, ext in extensions:
            for aggregate_name, options in ext.snapshot_bindings():
                migration_chain = SnapshotMigrationChain(options.migrations)
                _validate_snapshot_migration_target(aggregate_name, options.schema_version, migration_chain)
                configs[aggregate_name] = SnapshotConfig(
                    strategy=options.strategy,
                    schema_version=options.schema_version,
                    migration_chain=migration_chain,
                )
        return SnapshotConfigRegistry(configs)

    @classmethod
    def _build_type_registry(cls, aggregated: EventSourcingRegistry) -> tuple[EventTypeRegistry, UpcasterChain]:
        registry = EventTypeRegistry()
        upcasters: dict[str, Sequence[IEventUpcaster]] = {}

        for spec in cls._deduplicate(aggregated.event_type_bindings):
            item = spec if isinstance(spec, EventType) else EventType(spec)

            registry.register(item.event_type, name=item.name, version=item.version)
            for alias in item.aliases:
                registry.add_alias(item.event_type, alias)

            if item.upcasters:
                type_name = item.name or item.event_type.__name__
                cls._validate_upcaster_versions(item.upcasters, type_name, item.version)
                if type_name in upcasters and upcasters[type_name] is not item.upcasters:
                    msg = f'Conflicting upcaster definitions for event type {type_name!r}'
                    raise UpcasterChainError(msg)
                upcasters[type_name] = item.upcasters

        registry.freeze()
        return registry, UpcasterChain({k: list(v) for k, v in upcasters.items()})

    @staticmethod
    def _deduplicate(bindings: Sequence[EventTypeSpec]) -> Iterator[EventTypeSpec]:
        seen: set[int] = set()
        for item in bindings:
            item_id = id(item)
            if item_id not in seen:
                seen.add(item_id)
                yield item

    @staticmethod
    def _validate_upcaster_versions(upcasters: Sequence[IEventUpcaster], type_name: str, version: int) -> None:
        for u in upcasters:
            if u.from_version >= version:
                msg = (
                    f'Upcaster from_version {u.from_version} for event type {type_name!r} '
                    f'must be < event version {version}'
                )
                raise UpcasterChainError(msg)


def _validate_snapshot_migration_target(
    aggregate_name: str,
    schema_version: int,
    chain: SnapshotMigrationChain,
) -> None:
    if not chain.migrations:
        if schema_version != 1:
            msg = (
                f'Snapshot config for aggregate {aggregate_name!r}: '
                f'schema_version is {schema_version} but no migrations are provided. '
                f'Either set schema_version=1 or provide migrations.'
            )
            raise SnapshotMigrationChainError(msg)
        return

    first_from_version = chain.migrations[0].from_version
    if first_from_version != 1:
        msg = (
            f'Snapshot config for aggregate {aggregate_name!r}: '
            f'migration chain starts at version {first_from_version} but must start at version 1.'
        )
        raise SnapshotMigrationChainError(msg)

    final_to_version = chain.migrations[-1].to_version
    if final_to_version != schema_version:
        msg = (
            f'Snapshot config for aggregate {aggregate_name!r}: '
            f'migration chain reaches version {final_to_version} but schema_version is {schema_version}. '
            f'The final migration to_version must equal schema_version.'
        )
        raise SnapshotMigrationChainError(msg)



from waku.eventsourcing.decider.handler import DeciderCommandHandler, DeciderVoidCommandHandler
from waku.eventsourcing.decider.repository import DeciderRepository, SnapshotDeciderRepository

__all__ = [
    'DeciderCommandHandler',
    'DeciderRepository',
    'DeciderVoidCommandHandler',
    'SnapshotDeciderRepository',
]



from waku.eventsourcing.projection.binding import CatchUpProjectionBinding
from waku.eventsourcing.projection.checkpoint import Checkpoint
from waku.eventsourcing.projection.config import LeaseConfig, PollingConfig
from waku.eventsourcing.projection.in_memory import InMemoryCheckpointStore
from waku.eventsourcing.projection.interfaces import ErrorPolicy, ICatchUpProjection, ICheckpointStore, IProjection
from waku.eventsourcing.projection.registry import CatchUpProjectionRegistry
from waku.eventsourcing.projection.runner import CatchUpProjectionRunner

__all__ = [
    'CatchUpProjectionBinding',
    'CatchUpProjectionRegistry',
    'CatchUpProjectionRunner',
    'Checkpoint',
    'ErrorPolicy',
    'ICatchUpProjection',
    'ICheckpointStore',
    'IProjection',
    'InMemoryCheckpointStore',
    'LeaseConfig',
    'PollingConfig',
]



from __future__ import annotations

from waku.eventsourcing.contracts.stream import StreamId  # noqa: TC001  # used as dict key type
from waku.eventsourcing.snapshot.interfaces import ISnapshotStore, Snapshot

__all__ = ['InMemorySnapshotStore']


class InMemorySnapshotStore(ISnapshotStore):
    def __init__(self) -> None:
        self._snapshots: dict[StreamId, Snapshot] = {}

    async def load(self, stream_id: StreamId, /) -> Snapshot | None:
        return self._snapshots.get(stream_id)

    async def save(self, snapshot: Snapshot, /) -> None:
        self._snapshots[snapshot.stream_id] = snapshot



from __future__ import annotations

import abc
import logging
import typing
import uuid
from typing import ClassVar, Final, Generic, cast

from waku.eventsourcing._introspection import is_abstract, resolve_generic_args
from waku.eventsourcing._stream_helpers import read_aggregate_stream
from waku.eventsourcing.contracts.aggregate import (  # Dishka needs runtime access
    CommandT,
    EventT,
    IDecider,
    StateT,
)
from waku.eventsourcing.contracts.event import EventEnvelope, StoredEvent
from waku.eventsourcing.contracts.stream import Exact, NoStream, StreamId
from waku.eventsourcing.exceptions import StreamNotFoundError
from waku.eventsourcing.serialization.interfaces import (
    ISnapshotStateSerializer,  # noqa: TC001  # Dishka needs runtime access
)
from waku.eventsourcing.snapshot.interfaces import ISnapshotStore  # noqa: TC001  # Dishka needs runtime access
from waku.eventsourcing.snapshot.manager import SnapshotManager
from waku.eventsourcing.snapshot.registry import SnapshotConfigRegistry  # noqa: TC001  # Dishka needs runtime access
from waku.eventsourcing.store.interfaces import IEventStore  # noqa: TC001  # Dishka needs runtime access

__all__ = [
    'DeciderRepository',
    'SnapshotDeciderRepository',
]

logger = logging.getLogger(__name__)

_STATE_SUFFIX: Final = 'State'


class DeciderRepository(abc.ABC, Generic[StateT, CommandT, EventT]):
    aggregate_name: ClassVar[str]
    max_stream_length: ClassVar[int | None] = None

    def __init_subclass__(cls, **kwargs: object) -> None:
        super().__init_subclass__(**kwargs)
        if is_abstract(cls):
            return
        if not getattr(cls, 'aggregate_name', None):
            state_cls = cls._resolve_state_type()
            if state_cls is not None:
                name = state_cls.__name__
                if name.endswith(_STATE_SUFFIX) and len(name) > len(_STATE_SUFFIX):
                    name = name.removesuffix(_STATE_SUFFIX)
                cls.aggregate_name = name
            else:
                msg = f'{cls.__name__} must define aggregate_name or parametrize Generic with a concrete state type'
                raise TypeError(msg)

    @classmethod
    def _resolve_state_type(cls) -> type[StateT] | None:
        args = resolve_generic_args(cls, DeciderRepository)
        return args[0] if args else None  # ty: ignore[invalid-return-type]

    def __init__(
        self,
        decider: IDecider[StateT, CommandT, EventT],
        event_store: IEventStore,
    ) -> None:
        self._decider = decider
        self._event_store = event_store

    async def load(self, aggregate_id: str) -> tuple[StateT, int]:
        stream_id = self._stream_id(aggregate_id)
        stored_events = await read_aggregate_stream(
            self._event_store,
            stream_id,
            aggregate_name=self.aggregate_name,
            aggregate_id=aggregate_id,
            max_stream_length=self.max_stream_length,
        )
        state = self._decider.initial_state()
        for stored in cast('list[StoredEvent[EventT]]', stored_events):
            state = self._decider.evolve(state, stored.data)
        version = stored_events[-1].position if stored_events else -1
        logger.debug('Loaded %d events for %s/%s', len(stored_events), self.aggregate_name, aggregate_id)
        return state, version

    async def save(
        self,
        aggregate_id: str,
        events: typing.Sequence[EventT],
        expected_version: int,
        *,
        current_state: StateT | None = None,  # noqa: ARG002
        idempotency_key: str | None = None,
    ) -> int:
        if not events:
            return expected_version
        stream_id = self._stream_id(aggregate_id)
        envelopes = [
            EventEnvelope(
                domain_event=e,
                idempotency_key=f'{idempotency_key}:{i}' if idempotency_key else str(uuid.uuid4()),
            )
            for i, e in enumerate(events)
        ]
        expected = Exact(version=expected_version) if expected_version >= 0 else NoStream()
        new_version = await self._event_store.append_to_stream(stream_id, envelopes, expected_version=expected)
        logger.debug(
            'Saved %d events to %s/%s, version %d',
            len(events),
            self.aggregate_name,
            aggregate_id,
            new_version,
        )
        return new_version

    def _stream_id(self, aggregate_id: str) -> StreamId:
        return StreamId.for_aggregate(self.aggregate_name, aggregate_id)


class SnapshotDeciderRepository(DeciderRepository[StateT, CommandT, EventT], abc.ABC):
    snapshot_state_type: ClassVar[str | None] = None

    def __init__(
        self,
        decider: IDecider[StateT, CommandT, EventT],
        event_store: IEventStore,
        snapshot_store: ISnapshotStore,
        snapshot_config_registry: SnapshotConfigRegistry,
        state_serializer: ISnapshotStateSerializer,
    ) -> None:
        super().__init__(decider, event_store)
        self._state_serializer = state_serializer
        self._state_type: type[StateT] = type(self._decider.initial_state())
        config = snapshot_config_registry.get(self.aggregate_name)
        self._snapshot_manager = SnapshotManager(
            store=snapshot_store,
            config=config,
            state_type_name=self.snapshot_state_type or self._state_type.__name__,
        )

    async def load(self, aggregate_id: str) -> tuple[StateT, int]:
        stream_id = self._stream_id(aggregate_id)
        snapshot = await self._snapshot_manager.load_snapshot(stream_id, aggregate_id)

        if snapshot is not None:
            logger.debug('Loaded snapshot for %s/%s at version %d', self.aggregate_name, aggregate_id, snapshot.version)
            state = self._state_serializer.deserialize(snapshot.state, self._state_type)
            try:
                stored_events = await self._event_store.read_stream(stream_id, start=snapshot.version + 1)
            except StreamNotFoundError:
                stored_events = []
            for stored in cast('list[StoredEvent[EventT]]', stored_events):
                state = self._decider.evolve(state, stored.data)
            version = stored_events[-1].position if stored_events else snapshot.version
            return state, version

        logger.debug('No snapshot for %s/%s, loading from events', self.aggregate_name, aggregate_id)
        return await super().load(aggregate_id)

    async def save(
        self,
        aggregate_id: str,
        events: typing.Sequence[EventT],
        expected_version: int,
        *,
        current_state: StateT | None = None,
        idempotency_key: str | None = None,
    ) -> int:
        new_version = await super().save(
            aggregate_id,
            events,
            expected_version,
            current_state=current_state,
            idempotency_key=idempotency_key,
        )

        if events and self._snapshot_manager.should_save(aggregate_id, new_version):
            if current_state is not None:
                state = current_state
            else:
                state, _ = await self.load(aggregate_id)
            state_data = self._state_serializer.serialize(state)
            stream_id = self._stream_id(aggregate_id)
            await self._snapshot_manager.save_snapshot(stream_id, aggregate_id, state_data, new_version)

        return new_version



from __future__ import annotations

import abc
import logging
import uuid
from typing import TYPE_CHECKING, ClassVar, Generic, TypeVar

from waku.eventsourcing._introspection import is_abstract, resolve_generic_args
from waku.eventsourcing._stream_helpers import read_aggregate_stream
from waku.eventsourcing.contracts.aggregate import EventSourcedAggregate
from waku.eventsourcing.contracts.event import EventEnvelope
from waku.eventsourcing.contracts.stream import Exact, NoStream, StreamId
from waku.eventsourcing.exceptions import EventSourcingError
from waku.eventsourcing.store.interfaces import IEventStore  # noqa: TC001  # Dishka needs runtime access

if TYPE_CHECKING:
    from waku.cqrs.contracts.notification import INotification

__all__ = ['EventSourcedRepository']

logger = logging.getLogger(__name__)

AggregateT = TypeVar('AggregateT', bound=EventSourcedAggregate)


class EventSourcedRepository(abc.ABC, Generic[AggregateT]):
    aggregate_name: ClassVar[str]
    max_stream_length: ClassVar[int | None] = None

    def __init_subclass__(cls, **kwargs: object) -> None:
        super().__init_subclass__(**kwargs)
        if is_abstract(cls):
            return
        if not getattr(cls, 'aggregate_name', None):
            aggregate_cls = cls._resolve_aggregate_type()
            if aggregate_cls is not None:
                cls.aggregate_name = aggregate_cls.__name__
            else:
                msg = f'{cls.__name__} must define aggregate_name or parametrize Generic with a concrete type'
                raise TypeError(msg)

    @classmethod
    def _resolve_aggregate_type(cls) -> type[AggregateT] | None:
        args = resolve_generic_args(cls, EventSourcedRepository)
        return args[0] if args else None  # ty: ignore[invalid-return-type]

    def __init__(self, event_store: IEventStore) -> None:
        self._event_store = event_store

    async def load(self, aggregate_id: str) -> AggregateT:
        stream_id = self._stream_id(aggregate_id)
        stored_events = await read_aggregate_stream(
            self._event_store,
            stream_id,
            aggregate_name=self.aggregate_name,
            aggregate_id=aggregate_id,
            max_stream_length=self.max_stream_length,
        )
        if not stored_events:
            msg = f'Stream contains no events: {stream_id}'
            raise EventSourcingError(msg)
        aggregate = self.create_aggregate()
        domain_events = [e.data for e in stored_events]
        version = stored_events[-1].position
        logger.debug('Loaded %d events for %s/%s', len(stored_events), self.aggregate_name, aggregate_id)
        aggregate.load_from_history(domain_events, version)
        return aggregate

    async def save(
        self,
        aggregate_id: str,
        aggregate: AggregateT,
        *,
        idempotency_key: str | None = None,
    ) -> tuple[int, list[INotification]]:
        stream_id = self._stream_id(aggregate_id)
        events = aggregate.collect_events()
        if not events:
            return aggregate.version, []

        envelopes = [
            EventEnvelope(
                domain_event=event,
                idempotency_key=f'{idempotency_key}:{i}' if idempotency_key else str(uuid.uuid4()),
            )
            for i, event in enumerate(events)
        ]
        expected = Exact(version=aggregate.version) if aggregate.version >= 0 else NoStream()
        new_version = await self._event_store.append_to_stream(stream_id, envelopes, expected_version=expected)
        aggregate.mark_persisted(new_version)
        logger.debug(
            'Saved %d events to %s/%s, version %d',
            len(events),
            self.aggregate_name,
            aggregate_id,
            new_version,
        )
        return new_version, events

    def create_aggregate(self) -> AggregateT:
        aggregate_cls = self._resolve_aggregate_type()
        if aggregate_cls is None:
            msg = f'{type(self).__name__}: cannot auto-create aggregate, override create_aggregate()'
            raise TypeError(msg)
        return aggregate_cls()

    def _stream_id(self, aggregate_id: str) -> StreamId:
        return StreamId.for_aggregate(self.aggregate_name, aggregate_id)



from waku.eventsourcing.serialization.interfaces import ISnapshotStateSerializer
from waku.eventsourcing.serialization.json import JsonSnapshotStateSerializer
from waku.eventsourcing.snapshot.in_memory import InMemorySnapshotStore
from waku.eventsourcing.snapshot.interfaces import ISnapshotStore, ISnapshotStrategy, Snapshot
from waku.eventsourcing.snapshot.migration import ISnapshotMigration, SnapshotMigrationChain
from waku.eventsourcing.snapshot.registry import SnapshotConfig, SnapshotConfigRegistry
from waku.eventsourcing.snapshot.repository import SnapshotEventSourcedRepository
from waku.eventsourcing.snapshot.strategy import EventCountStrategy

__all__ = [
    'EventCountStrategy',
    'ISnapshotMigration',
    'ISnapshotStateSerializer',
    'ISnapshotStore',
    'ISnapshotStrategy',
    'InMemorySnapshotStore',
    'JsonSnapshotStateSerializer',
    'Snapshot',
    'SnapshotConfig',
    'SnapshotConfigRegistry',
    'SnapshotEventSourcedRepository',
    'SnapshotMigrationChain',
]



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence

    from waku.eventsourcing.contracts.event import EventMetadata, IMetadataEnricher

__all__ = ['enrich_metadata']


def enrich_metadata(
    metadata: EventMetadata,
    enrichers: Sequence[IMetadataEnricher],
) -> EventMetadata:
    for enricher in enrichers:
        metadata = enricher.enrich(metadata)
    return metadata
